/*
** Prototype RasterImport module for importing TIFF images into
** Xew internal raster format.
**
** Copyright 1992, 1993 by Markku Savela and
**	Technical Research Centre of Finland
**
** The author may be contacted via:
**
**    Mail:	Markku Savela
**		Technical Research Centre of Finland
**		Telecommunications Laboratory
**		Otakaari 7 B, P.O. Box 34
**		SF-02151 ESPOO
**		Finland
**
**    Phone:	+358 0 456 6509
**    Email:	savela@tel.vtt.fi
*/

#if SYSV_INCLUDES
#	include <malloc.h>	/* malloc() et al. */
#endif
#if ANSI_INCLUDES
#	include <stdlib.h>
#endif
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xew/RasterP.h>
#include "RasterTIFF.h"

#ifndef NO_TIFF
#	include "tiffio.h"
#endif

static void RasterImportWarning(w, msg)
XeRasterWidget w;
char *msg;
    {
	XeWidgetWarningMsg
		((Widget)w, "rasterImportWarning", msg, (String *)NULL, 0);
    }


void XeImport_TIFF(w)
XeRasterWidget w;
    {
#ifdef NO_TIFF
	RasterImportWarning(w, "TIFF Support not compiled.");
#else
	TIFF *in;
	unsigned long width, height;
	unsigned short *red, *green, *blue;
	unsigned short samplesperpixel, bitspersample;
	unsigned short photometric, orientation;
	float resX, resY;
	unsigned short resU;
	register unsigned long row;
	unsigned char *data;
	int linesize;

	if ((in = TIFFOpen(w->basic.content_file, "r")) == NULL)
	    {
		XeWidgetWarningMsg
			((Widget)w,
			 "rasterImportWarning",
			 "Could not open TIFF %s",
			 &w->basic.content_file,
			 1);
		return;
	    }
	TIFFGetFieldDefaulted(in, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
	w->raster.samples_per_pixel = samplesperpixel;
	TIFFGetFieldDefaulted(in, TIFFTAG_BITSPERSAMPLE, &bitspersample);
	w->raster.bits_per_sample = bitspersample;
	w->raster.bits_per_component = bitspersample;
	/*
	** Load Colormap if defined, regardless whether photometric
	** indicates it is needed or not.
	*/
	if (TIFFGetField(in, TIFFTAG_COLORMAP, &red, &green, &blue))
	    {
		int mapsize = (1<<bitspersample);
		unsigned char *r, *g, *b;
		int x;

		w->raster.color_map = (unsigned char *)XtMalloc(3 * mapsize);
		/*
		** Convert 16-bit colormap to 8-bit colormap
		*/
		r = w->raster.color_map;
		g = r + mapsize;
		b = g + mapsize;
		for (x = mapsize; --x >= 0; )
		    {
			r[x] = (red[x] * 255) / 65535;
			g[x] = (green[x] * 255) / 65535;
			b[x] = (blue[x] * 255) / 65535;
		    }
	    }
	if (TIFFGetField(in, TIFFTAG_PHOTOMETRIC, &photometric) == 0)
	    {
		/*
		** Missing PHOTOMETRIC.. faulty TIFF. Make some
		** guess so that at least something will be shown.
		*/
		RasterImportWarning(w, "PHOTOMETRIC tag missing from TIFF");
		if (samplesperpixel > 1)
			photometric = PHOTOMETRIC_RGB;
		else if (w->raster.color_map != NULL)
			photometric = PHOTOMETRIC_PALETTE;
		else
			photometric = PHOTOMETRIC_MINISBLACK;
			
	    }
	TIFFGetFieldDefaulted(in, TIFFTAG_ORIENTATION, &orientation);
	if (orientation != ORIENTATION_TOPLEFT)
		XeWidgetWarningMsg
			((Widget)w,
			 "rasterImportWarning",
			 "Warning orientation=%d",	/* uhh.. --msa */
			 (String *)&orientation,	/* ..... --msa */
			 1);
	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &width);
	w->raster.width = width;
	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &height);
	w->raster.height = height;
	if (TIFFGetField(in, TIFFTAG_XRESOLUTION, &resX))
		w->raster.x_resolution = resX;
	else
		w->raster.x_resolution = 1.0;
	if (TIFFGetField(in, TIFFTAG_YRESOLUTION, &resY))
		w->raster.y_resolution = resY;
	else
		w->raster.y_resolution = 1.0;
	TIFFGetFieldDefaulted(in, TIFFTAG_RESOLUTIONUNIT, &resU);
	w->raster.resolution_unit =
		resU == RESUNIT_NONE ? XeResolutionUnit_NONE :
		resU == RESUNIT_INCH ? XeResolutionUnit_INCH :
		resU == RESUNIT_CENTIMETER ? XeResolutionUnit_CM :
			XeResolutionUnit_NONE;
	w->raster.bytes_per_line = linesize = TIFFScanlineSize(in);
	data = (unsigned char *)malloc(linesize * height);
	if (data != NULL)
	    {
		w->raster.data = data;
		w->raster.copy = TRUE;
		for (row = 0; row < height; row++)
		    {
			if (TIFFReadScanline(in, data, row, 0) < 0)
				break;
			data += linesize;
		    }
		if (photometric == PHOTOMETRIC_MINISBLACK &&
		    bitspersample == 1)
		    {
			/*
			** If 0 = Black, we have to invert the image
			** (this works only for bilevel, fix grayscale
			** later --msa)
			*/
			row = linesize * height;
			data = w->raster.data;
			for ( ; row > 0; row -= 1, data += 1)
				*data = ~*data;
		    }
	    }
	switch (photometric)
	    {
	    case PHOTOMETRIC_MINISWHITE:
	    case PHOTOMETRIC_MINISBLACK:
		if (bitspersample == 1)
			w->raster.class = XeRasterClass_BILEVEL;
		else
		    {
			/*
			** A Grayscale Image, make it palette colour
			*/
			w->raster.class = XeRasterClass_PALETTE;
			if (w->raster.color_map == NULL)
			    {
				int mapsize = (1<<bitspersample);
				unsigned char *r, *g, *b;
				int x, i, v;
				
				w->raster.color_map =
					(unsigned char *)XtMalloc(3 * mapsize);
				/*
				** Create a dummy colormap
				*/
				r = w->raster.color_map;
				g = r + mapsize;
				b = g + mapsize;
				x = 256 >> bitspersample;
				if (photometric == PHOTOMETRIC_MINISBLACK)
					v = 0;
				else
				    {
					v = 255;
					x = -x;
				    }
				for (i = 0 ; i < mapsize; i++, v += x)
					r[i] = g[i] = b[i] = v;
			    }
		    }
		break;
	    case PHOTOMETRIC_RGB:
		w->raster.class = XeRasterClass_FULLCOLOR;
		break;
	    case PHOTOMETRIC_PALETTE:
		w->raster.class = XeRasterClass_PALETTE;
		break;
	    default:
		w->raster.class = XeRasterClass_UNKNOWN;
		break;
	    }
	TIFFClose(in);
#endif
    }
